package org.jboss.soa.esb.samples.quickstart.contract_publisher_soapprocessor_import;

import java.io.StringReader;
import java.io.StringWriter;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.dom4j.Attribute;
import org.dom4j.Document;
import org.dom4j.Element;
import org.dom4j.io.OutputFormat;
import org.dom4j.io.SAXReader;
import org.dom4j.io.XMLWriter;
import org.jaxen.SimpleNamespaceContext;
import org.jaxen.XPath;
import org.jaxen.dom4j.Dom4jXPath;

public class ContractPublisherSOAPProcessorImportTest 
{
	private static final Logger logger = Logger.getLogger(ContractPublisherSOAPProcessorImportTest.class);
	private static final String CATEGORY = "ContractPublisherSOAPProducerCategory";
	private static final String SERVICE = "ContractPublisherSOAPProducerImportService";
	
	// expected republished endpoint addresses
	private static final String EXPECTED_JMS_ENDPOINT   = "jms:/#queue/contract_publisher_soapproducer_import_esb";
	private static final String EXPECTED_HTTP_ENDPOINT  = "http://localhost:8765/";
	private static final String EXPECTED_SOCKET_ENDPOINT= "socket://localhost:8888/";
	
	// WSDL document of the original service 
	private static final String ORIGINAL_URL = "http://localhost:8080/Contract_publisher_soapproducer_import/GoodbyeWorldWS?wsdl";
	
	// URLs for the republished endpoints
	private static final String CONTRACT_URL = "http://localhost:8080/contract/contract.jsp?serviceCat="+CATEGORY+"&serviceName="+SERVICE+"&protocol=";
	private static final String JMS_URL    = CONTRACT_URL + "jms";
	private static final String HTTP_URL   = CONTRACT_URL + "http";
	private static final String SOCKET_URL = CONTRACT_URL + "socket";
	
	// XPaths statements used for rewritting WSDLs
	private static final String XPATH_BINDING = "//soap:binding/@transport";
	private static final String XPATH_ADDRESS = "//soap:address/@location";
	private static final String XPATH_IMPORT  = "//wsdl:import/@location"; 
	
	public void testImports() throws Exception
	{
		logger.info("Entering test method.");
		logger.info("Comparing WSDL of JMS republished endpoint.");
		compareWSDLDocuments(ORIGINAL_URL, JMS_URL, EXPECTED_JMS_ENDPOINT);
		logger.info("Comparing republished WSDL for the HTTP endpoint.");
		compareWSDLDocuments(ORIGINAL_URL, HTTP_URL, EXPECTED_HTTP_ENDPOINT);
		logger.info("Comparing republished WSDL for SOCKET endpoint.");
		compareWSDLDocuments(ORIGINAL_URL, SOCKET_URL, EXPECTED_SOCKET_ENDPOINT);
		logger.info("Leaving test method.");
	}
	
	private boolean compareImports(String importUrl) throws Exception 
	{
		logger.info("Reading imported types from: "+ importUrl);
		Document importedTypes = openDocument(importUrl);
		Document originalTypes = new SAXReader().read(getClass().getResourceAsStream("/GoodByeSchema.xsd"));
		return toXMLString(getSchemas(importedTypes)).equals(toXMLString(getSchemas(originalTypes)));
	}
	
	private void compareWSDLDocuments(String originalUrl, String proxiedUrl, String expectedEndpoint) throws Exception 
	{
		logger.info("Reading original WSDL from: "+ originalUrl);
		Document original = openDocument(originalUrl);
		logger.info("Reading republished WSDL from: "+ proxiedUrl);
		Document proxied = openDocument(proxiedUrl);
		logger.info("Asserting operation names in WSDL: "+ getOperationsAsList(proxied).equals(getOperationsAsList(original)));
		logger.info("Asserting location attribute value from WSDL: "+ getLocationValue(proxied, XPATH_ADDRESS).equals(expectedEndpoint));
		
		logger.info("Asserting imported types.");
		logger.info("Comparison of imported types should be true: "+ compareImports(getLocationValue(proxied, XPATH_IMPORT)));
		logger.info("Rewritting documents to make comparison easier.");
		rewrite(original, XPATH_ADDRESS);
		rewrite(original, XPATH_BINDING);
		rewrite(original, XPATH_IMPORT);
		
		rewrite(proxied, XPATH_ADDRESS);
		rewrite(proxied, XPATH_BINDING);
		rewrite(proxied, XPATH_IMPORT);
		logger.info("Comparing documents as XML strings: "+ toXMLString(original).equals(toXMLString(proxied)));
	}
	
	private Document openDocument(String url) throws Exception 
	{
		return new SAXReader().read(new URL(url));
	}
	
	private void rewrite(Document doc, String xpathExpr) throws Exception
	{
		
		Map map = new HashMap();		
		map.put("wsdl", "http://schemas.xmlsoap.org/wsdl/");
		map.put("soap", "http://schemas.xmlsoap.org/wsdl/soap/");
		XPath xpath = new Dom4jXPath(xpathExpr);
		xpath.setNamespaceContext(new SimpleNamespaceContext(map));
		List list = xpath.selectNodes(doc);
		
			Attribute attr = (Attribute) list.get(0);
			attr.setValue("####");	
	}
	
	private String toXMLString(Element input) throws Exception
	{
		StringWriter sw = new StringWriter();
		XMLWriter writer = new XMLWriter(sw, OutputFormat.createCompactFormat());
		writer.write(input);
		return new SAXReader().read(new StringReader(sw.toString())).asXML();
	}
	
	private String toXMLString(Document input) throws Exception
	{
		StringWriter sw = new StringWriter();
		XMLWriter writer = new XMLWriter(sw, OutputFormat.createCompactFormat());
		writer.write(input);
		return new SAXReader().read(new StringReader(sw.toString())).asXML();
	}
	
	private List<String> getOperationsAsList(Document wsdl) throws Exception
	{
		Map map = new HashMap();
		List<String> ops = new ArrayList<String>();
		
		map.put("wsdl", "http://schemas.xmlsoap.org/wsdl/");
		XPath xpath = new Dom4jXPath("//wsdl:binding/wsdl:operation/@name");
		xpath.setNamespaceContext(new SimpleNamespaceContext(map));
		List list = xpath.selectNodes(wsdl);
		
		for (int i = 0; i < list.size(); i++)
		{
			Attribute attr = (Attribute) list.get(i);
			ops.add(attr.getValue());
		}
		return ops;
	}
		
	private Element getSchemas(Document doc) throws  Exception 
	{
		Map map = new HashMap();
		map.put("xs", "http://www.w3.org/2001/XMLSchema");
		XPath xpath = new Dom4jXPath("//xs:schema");
		xpath.setNamespaceContext(new SimpleNamespaceContext(map));
		List list = xpath.selectNodes(doc);
		return ((Element) list.get(0) );
	}
	
	private Document readReasourceAsDocument(String resourceName) throws Exception 
	{
		return new SAXReader().read(getClass().getResourceAsStream(resourceName));
	}
	
	private String getLocationValue(Document doc, String xpathExpr) throws Exception
	{
		Map map = new HashMap();
		map.put("wsdl", "http://schemas.xmlsoap.org/wsdl/");
		map.put("soap", "http://schemas.xmlsoap.org/wsdl/soap/");
		XPath xpath = new Dom4jXPath(xpathExpr);
		xpath.setNamespaceContext(new SimpleNamespaceContext(map));
		List list = xpath.selectNodes(doc);
		
		return ((Attribute) list.get(0)).getValue();
	}
	
	public static void main(String[] args) throws Exception 
	{
		ContractPublisherSOAPProcessorImportTest test = new ContractPublisherSOAPProcessorImportTest();
		test.testImports();
	}
}